// Parameter settings for effect sliders: Default, Minimum, Maximum, Increment.
// 0.25  0.0  0.5  0.02
// 0.50  0.0  1.0  0.02

uniform sampler2D sampler;      // Texture sampler
uniform float param1;           // Effect parameter 1 changable via slider in the GUI
uniform float param2;           // Effect parameter 2 changable via slider in the GUI
uniform float texScaleS;        // Texture scale in S for non-quadratic images
uniform float texScaleT;        // Texture scale in T for non-quadratic images

void main(void)
{
    vec2 texCoord = gl_TexCoord[0].xy;      // [0.0 ,1.0] x [0.0, 1.0]
    texCoord.x /= texScaleS;
    texCoord.y /= texScaleT;
    vec2 normCoord = 2.0 * texCoord - 1.0;  // [-1.0 ,1.0] x [-1.0, 1.0]

    // Effect function: Stretch
    vec2 s = sign(normCoord);
    normCoord = abs(normCoord);
    normCoord = 0.5 * normCoord + 0.5 * smoothstep(param1, param2, normCoord) * normCoord;
    normCoord = s * normCoord;

    texCoord = normCoord / 2.0 + 0.5; // [0.0 ,1.0] x [0.0, 1.0]
    texCoord.x *= texScaleS;
    texCoord.y *= texScaleT;

    gl_FragColor = texture2D(sampler, texCoord);
}
